! ***********************************************************************
!
!   Copyright (C) 2010-2019  Bill Paxton & The MESA Team
!
!   this file is part of mesa.
!
!   mesa is free software; you can redistribute it and/or modify
!   it under the terms of the gnu general library public license as published
!   by the free software foundation; either version 2 of the license, or
!   (at your option) any later version.
!
!   mesa is distributed in the hope that it will be useful, 
!   but without any warranty; without even the implied warranty of
!   merchantability or fitness for a particular purpose.  see the
!   gnu library general public license for more details.
!
!   you should have received a copy of the gnu library general public license
!   along with this software; if not, write to the free software
!   foundation, inc., 59 temple place, suite 330, boston, ma 02111-1307 usa
!
! ***********************************************************************
 
      module run_star_extras

      use star_lib
      use star_def
      use const_def
      use math_lib
      use chem_def, only: chem_isos
      
      implicit none
      
      ! these routines are called by the standard run_star check_model
      contains
      
      subroutine extras_controls(id, ierr)
         integer, intent(in) :: id
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         
         ! this is the place to set any procedure pointers you want to change
         ! e.g., other_wind, other_mixing, other_energy  (see star_data.inc)


         ! the extras functions in this file will not be called
         ! unless you set their function pointers as done below.
         ! otherwise we use a null_ version which does nothing (except warn).

         s% other_energy => entropy_injection_routine
         s% extras_startup => extras_startup
         s% extras_start_step => extras_start_step
         s% extras_check_model => extras_check_model
         s% extras_finish_step => extras_finish_step
         s% extras_after_evolve => extras_after_evolve
         s% how_many_extra_history_columns => how_many_extra_history_columns
         s% data_for_extra_history_columns => data_for_extra_history_columns
         s% how_many_extra_profile_columns => how_many_extra_profile_columns
         s% data_for_extra_profile_columns => data_for_extra_profile_columns  

         s% how_many_extra_history_header_items => how_many_extra_history_header_items
         s% data_for_extra_history_header_items => data_for_extra_history_header_items
         s% how_many_extra_profile_header_items => how_many_extra_profile_header_items
         s% data_for_extra_profile_header_items => data_for_extra_profile_header_items

      end subroutine extras_controls


      subroutine entropy_injection_routine(id, ierr)
         use const_def, only: Msun
         integer, intent(in) :: id
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         integer :: k, k_core, He, C, O
	 double precision :: trel, mtransition, etatransition, highentropy, lowentropy, tempWD, alpha, s_targ, m_core, m_shell

         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return

	 ! default everything to zero
         s% extra_heat(:) = 0d0
         s% d_extra_heat_dlnd00(:) = 0d0
         s% d_extra_heat_dlnT00(:) = 0d0

         ! set the timescale to relax the entropy profile to be short,
         ! so that there isn't time for thermal transport
         ! or convection to mess up the desired profile
         trel = s% x_ctrl(5)

         ! loop through the zones and relax the
         ! entropy profile to the desired values
         ! this particular parametrization is definitely not fundamental
         ! it's just meant to mimic the entropy equation in such a way
         ! that the extra heating or cooling has a vaguely appropriate magnitude
	 highentropy = s%x_ctrl(2)
         mtransition = s%x_ctrl(1)*Msun
         lowentropy = s%x_ctrl(3)
         etatransition = s%x_ctrl(4)
	 tempWD = s% x_ctrl(6) !3.325d7
	 alpha = s% x_ctrl(7)
	 m_core = s% x_ctrl(8)*Msun
	 m_shell = s% x_ctrl(9)*Msun
         !Find index for specific species
	 do k = 1, s%species, 1
	    if (chem_isos% name(s% chem_id(k)) == "he4") then
	       	He = k
	    end if
	    if (chem_isos% name(s% chem_id(k)) == "c12") then
	       	C = k
	    end if
	    if (chem_isos% name(s% chem_id(k)) == "o16") then
	       	O = k
	    end if
	 end do
	 !Find index of core as defined by chemical boundary
	 do k = 1, s%nz
	    !if (s%m(k) < m_core) then
	    !if (s% xa(He,k) .lt. s% xa(C,k)+s% xa(O,k)) then
            if (s% xa(He,k) .lt. .99*s%xa(He,1)) then
            !if (s% xa(He,k) .lt. 0.01) then
	          k_core = k
	      exit
	    end if
	 end do

	 m_core = s%m(k_core)
	 lowentropy = exp(s%lnS(k_core))

         do k = 1, s% nz
	    ! Setting the entropy profile to a Sigmoid function with variable step steepness, "alpha"
	    !s_targ = (highentropy-lowentropy) / (1+exp(-alpha*(s%m(k)-mtransition)/Msun))+lowentropy
            	    !s% extra_heat(k) = ( s_targ - exp(s%lnS(k)) ) * s%T(k)/trel
	    !if (s% m(k) .le. m_core) then
	       !s% extra_heat(k) = ( lowentropy - exp(s%lnS(k)) ) * s%T(k)/trel
	    !end if

	    if (s% m(k) .ge. m_core .and. s% m(k) .le. m_shell) then
	       s% extra_heat(k) = ( ((exp(s%lnS(k_core))-highentropy)/(m_core-m_shell))*(s%m(k)-m_core)+lowentropy - exp(s%lnS(k)) ) * s%T(k)/trel
	    end if

	    if (s% m(k) .gt. m_shell) then
	       s% extra_heat(k) = ( highentropy - exp(s%lnS(k)) ) * s%T(k)/trel
	    end if

            ! when the inner region becomes degenerate "enough" (eta > etatransition),
            ! make it isothermal, with temperature = tempWD
            if (s% m(k) .lt. m_core .and. s%eta(k) .ge. etatransition) then
                     s% extra_heat(k) = ( tempWD - s%T(k) )*exp(s%lnS(k))/trel
                     !s% extra_heat(k) = ( s%T(k_core+1) - s%T(k) )*exp(s%lnS(k))/trel
            end if
         end do
      end subroutine entropy_injection_routine
      
      
      subroutine extras_startup(id, restart, ierr)
         integer, intent(in) :: id
         logical, intent(in) :: restart
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
      end subroutine extras_startup
      

      integer function extras_start_step(id)
         integer, intent(in) :: id
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_start_step = 0
      end function extras_start_step


      ! returns either keep_going, retry, or terminate.
      integer function extras_check_model(id)
         integer, intent(in) :: id
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_check_model = keep_going         
         if (.false. .and. s% star_mass_h1 < 0.35d0) then
            ! stop when star hydrogen mass drops to specified level
            extras_check_model = terminate
            write(*, *) 'have reached desired hydrogen mass'
            return
         end if


         ! if you want to check multiple conditions, it can be useful
         ! to set a different termination code depending on which
         ! condition was triggered.  MESA provides 9 customizeable
         ! termination codes, named t_xtra1 .. t_xtra9.  You can
         ! customize the messages that will be printed upon exit by
         ! setting the corresponding termination_code_str value.
         ! termination_code_str(t_xtra1) = 'my termination condition'

         ! by default, indicate where (in the code) MESA terminated
         if (extras_check_model == terminate) s% termination_code = t_extras_check_model
      end function extras_check_model


      integer function how_many_extra_history_columns(id)
         integer, intent(in) :: id
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         if (s% x_logical_ctrl(1)) then
            how_many_extra_history_columns = s% species
         else
            how_many_extra_history_columns = 0
         end if
      end function how_many_extra_history_columns
      
      
      subroutine data_for_extra_history_columns(id, n, names, vals, ierr)
         integer, intent(in) :: id, n
         character (len=maxlen_history_column_name) :: names(n)
         real(dp) :: vals(n)
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
	 integer :: i, surf
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         
         ! note: do NOT add the extras names to history_columns.list
         ! the history_columns.list is only for the built-in history column options.
         ! it must not include the new column names you are adding here.
         if (s% x_logical_ctrl(1)) then
            !Find tau = 1
            do i=1,s% nz, 1
               surf = i
               if (s% tau(i) .ge. 1) exit
            end do
         end if

            !Calculate surface for each iso
            do i=1, s% species, 1
               names(i) = "rcb_"//chem_isos% name(s% chem_id(i))
               vals(i) = dot_product(s% xa(i,1:surf), s% dm(1:surf)) / sum(s% dm(1:surf))
            end do

      end subroutine data_for_extra_history_columns

      
      integer function how_many_extra_profile_columns(id)
         integer, intent(in) :: id
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         how_many_extra_profile_columns = 0
      end function how_many_extra_profile_columns
      
      
      subroutine data_for_extra_profile_columns(id, n, nz, names, vals, ierr)
         integer, intent(in) :: id, n, nz
         character (len=maxlen_profile_column_name) :: names(n)
         real(dp) :: vals(nz,n)
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         integer :: k
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         
         ! note: do NOT add the extra names to profile_columns.list
         ! the profile_columns.list is only for the built-in profile column options.
         ! it must not include the new column names you are adding here.

         ! here is an example for adding a profile column
         !if (n /= 1) stop 'data_for_extra_profile_columns'
         !names(1) = 'beta'
         !do k = 1, nz
         !   vals(k,1) = s% Pgas(k)/s% P(k)
         !end do
         
      end subroutine data_for_extra_profile_columns


      integer function how_many_extra_history_header_items(id)
         integer, intent(in) :: id
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         how_many_extra_history_header_items = 0
      end function how_many_extra_history_header_items


      subroutine data_for_extra_history_header_items(id, n, names, vals, ierr)
         integer, intent(in) :: id, n
         character (len=maxlen_history_column_name) :: names(n)
         real(dp) :: vals(n)
         type(star_info), pointer :: s
         integer, intent(out) :: ierr
         ierr = 0
         call star_ptr(id,s,ierr)
         if(ierr/=0) return

         ! here is an example for adding an extra history header item
         ! also set how_many_extra_history_header_items
         ! names(1) = 'mixing_length_alpha'
         ! vals(1) = s% mixing_length_alpha

      end subroutine data_for_extra_history_header_items


      integer function how_many_extra_profile_header_items(id)
         integer, intent(in) :: id
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         how_many_extra_profile_header_items = 0
      end function how_many_extra_profile_header_items


      subroutine data_for_extra_profile_header_items(id, n, names, vals, ierr)
         integer, intent(in) :: id, n
         character (len=maxlen_profile_column_name) :: names(n)
         real(dp) :: vals(n)
         type(star_info), pointer :: s
         integer, intent(out) :: ierr
         ierr = 0
         call star_ptr(id,s,ierr)
         if(ierr/=0) return

         ! here is an example for adding an extra profile header item
         ! also set how_many_extra_profile_header_items
         ! names(1) = 'mixing_length_alpha'
         ! vals(1) = s% mixing_length_alpha

      end subroutine data_for_extra_profile_header_items


      ! returns either keep_going or terminate.
      ! note: cannot request retry; extras_check_model can do that.
      integer function extras_finish_step(id)
         integer, intent(in) :: id
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_finish_step = keep_going

         ! to save a profile, 
            ! s% need_to_save_profiles_now = .true.
         ! to update the star log,
            ! s% need_to_update_history_now = .true.

         ! see extras_check_model for information about custom termination codes
         ! by default, indicate where (in the code) MESA terminated
         if (extras_finish_step == terminate) s% termination_code = t_extras_finish_step
      end function extras_finish_step
      
      
      subroutine extras_after_evolve(id, ierr)
         integer, intent(in) :: id
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
      end subroutine extras_after_evolve

      end module run_star_extras
      
